@@ -6,7 +6,7 @@ Mojo::SNMP - Run SNMP requests with Mojo::IOLoop
 
 =head1 VERSION
 
-0.10
+0.11
 
 =head1 SYNOPSIS
 
@@ -90,7 +90,7 @@ use Scalar::Util ();
 use constant DEBUG => $ENV{MOJO_SNMP_DEBUG} ? 1 : 0;
 use constant MAXREPETITIONS => 10;
 
-our $VERSION = '0.10';
+our $VERSION = '0.11';
 
 my $DISPATCHER;
 my @EXCLUDE_METHOD_ARGS = qw( maxrepetitions );
@@ -101,76 +101,8 @@ my %EXCLUDE             = (
 );
 
 my %SNMP_METHOD;
-__PACKAGE__->add_custom_request_method(
-  bulk_walk => sub {
-    my ($session, %args) = @_;
-    my $base_oid       = $args{varbindlist}[0];
-    my $last           = $args{callback};
-    my $maxrepetitions = $args{maxrepetitions} || MAXREPETITIONS;
-    my ($callback, $end, %tree, %types);
-
-    $end = sub {
-      $session->{_pdu}->var_bind_list(\%tree, \%types) if %tree;
-      $session->$last;
-      $end = $callback = undef;
-    };
-
-    $callback = sub {
-      my ($session) = @_;
-      my $res     = $session->var_bind_list    or return $end->();
-      my @sortres = $session->var_bind_names() or return $end->();
-      my $types   = $session->var_bind_types;
-      my $next    = $sortres[-1];
-
-      for my $oid (@sortres) {
-        return $end->() unless Net::SNMP::oid_base_match($base_oid, $oid);
-        $types{$oid} = $types->{$oid};
-        $tree{$oid}  = $res->{$oid};
-      }
-
-      return $end->() unless $next;
-      return $session->get_bulk_request(maxrepetitions => $maxrepetitions, varbindlist => [$next],
-        callback => $callback);
-    };
-
-    $session->get_bulk_request(maxrepetitions => $maxrepetitions, varbindlist => [$base_oid], callback => $callback);
-  }
-);
-
-__PACKAGE__->add_custom_request_method(
-  walk => sub {
-    my ($session, %args) = @_;
-    my $base_oid = $args{varbindlist}[0];
-    my $last     = $args{callback};
-    my ($callback, $end, %tree, %types);
-
-    $end = sub {
-      $session->{_pdu}->var_bind_list(\%tree, \%types) if %tree;
-      $session->$last;
-      $end = $callback = undef;
-    };
-
-    $callback = sub {
-      my ($session) = @_;
-      my $res = $session->var_bind_list or return $end->();
-      my $types = $session->var_bind_types;
-      my @next;
-
-      for my $oid (keys %$res) {
-        if (Net::SNMP::oid_base_match($base_oid, $oid)) {
-          $types{$oid} = $types->{$oid};
-          $tree{$oid}  = $res->{$oid};
-          push @next, $oid;
-        }
-      }
-
-      return $end->() unless @next;
-      return $session->get_next_request(varbindlist => \@next, callback => $callback);
-    };
-
-    $session->get_next_request(varbindlist => [$base_oid], callback => $callback);
-  }
-);
+__PACKAGE__->add_custom_request_method(bulk_walk => \&_snmp_method_bulk_walk);
+__PACKAGE__->add_custom_request_method(walk      => \&_snmp_method_walk);
 
 $Net::SNMP::DISPATCHER = $Net::SNMP::DISPATCHER;    # avoid warning
 
@@ -254,11 +186,32 @@ has ioloop         => sub { Mojo::IOLoop->singleton };
 # these attributes are experimental and therefore not exposed. Let me know if
 # you use them...
 has _dispatcher => sub { $DISPATCHER ||= Mojo::SNMP::Dispatcher->new(ioloop => shift->ioloop) };
-has _pool       => sub { +{} };
-has _queue      => sub { +[] };
 
 =head1 METHODS
 
+=head2 add_custom_request_method
+
+  $self->add_custom_request_method(name => sub {
+    my($session, %args) = @_;
+    # do custom stuff..
+  });
+
+This method can be used to add custom L<Net::SNMP> request methods. See the
+source code for an example on how to do "walk".
+
+NOTE: This method will also replace any method, meaning the code below will
+call the custom callback instead of L<Net::SNMP/get_next_request>.
+
+  $self->add_custom_request_method(get_next => $custom_callback);
+
+=cut
+
+sub add_custom_request_method {
+  my ($class, $name, $cb) = @_;
+  $SNMP_METHOD{$name} = $cb;
+  $class;
+}
+
 =head2 get
 
   $self->get($host, $args, \@oids, sub {
@@ -289,39 +242,6 @@ L</response> event. C<$args> is optional.
 Will call the callback when data is retrieved, instead of emitting the
 L</response> event. C<$args> is optional.
 
-=head2 set
-
-  $self->set($host, $args => [ $oid => OCTET_STRING, $value, ... ], sub {
-    my($self, $err, $res) = @_;
-    # ...
-  });
-
-Will call the callback when data is set, instead of emitting the
-L</response> event. C<$args> is optional.
-
-=head2 walk
-
-  $self->walk($host, $args, \@oids, sub {
-    my($self, $err, $res) = @_;
-    # ...
-  });
-
-Will call the callback when data is retrieved, instead of emitting the
-L</response> event. C<$args> is optional.
-
-=cut
-
-for my $method (qw( get get_bulk get_next set walk )) {
-  eval <<"HERE" or die $@;
-    sub $method {
-      my(\$self, \$host) = (shift, shift);
-      my \$args = ref \$_[0] eq 'HASH' ? shift : {};
-      \$self->prepare(\$host, \$args, $method => \@_);
-    }
-    1;
-HERE
-}
-
 =head2 prepare
 
   $self = $self->prepare($host, \%args, ...);
@@ -381,7 +301,7 @@ sub prepare {
   my $args  = ref $_[0] eq 'HASH' ? shift : {};
   my %args  = %$args;
 
-  $hosts = [sort keys %{$self->_pool}] if $hosts->[0] and $hosts->[0] eq '*';
+  $hosts = [keys %{$self->{sessions} || {}}] if $hosts->[0] and $hosts->[0] eq '*';
 
   defined $args{$_} or $args{$_} = $self->defaults->{$_} for keys %{$self->defaults};
   $args{version} = $self->_normalize_version($args{version} || '');
@@ -393,105 +313,46 @@ HOST:
     my ($host) = $key =~ /^([^|]+)/;
     local $args{hostname} = $host;
     my $key = $key eq $host ? $self->_calculate_pool_key(\%args) : $key;
-    $self->_pool->{$key} ||= $self->_new_session(\%args) or next HOST;
+    $self->{sessions}{$key} ||= $self->_new_session(\%args) or next HOST;
 
     local @_ = @_;
     while (@_) {
       my $method = shift;
       my $oid = ref $_[0] eq 'ARRAY' ? shift : [shift];
-      push @{$self->_queue}, [$key, $method, $oid, $args, $cb];
+      push @{$self->{queue}{$key}}, [$key, $method, $oid, $args, $cb];
     }
   }
 
-  $self->{_requests} ||= 0;
-  $self->_prepare_request or last for $self->{_requests} .. $self->concurrent - 1;
+  $self->{n_requests} ||= 0;
+
+  for ($self->{n_requests} .. $self->concurrent - 1) {
+    my $queue = $self->_dequeue or last;
+    $self->_prepare_request($queue);
+  }
+
   $self->_setup if !$self->{_setup}++ and $self->ioloop->is_running;
   $self;
 }
 
-sub _calculate_pool_key {
-  join '|', map { defined $_[1]->{$_} ? $_[1]->{$_} : '' } qw( hostname version community username );
-}
+=head2 set
 
-sub _normalize_version {
-  $_[1] =~ /1/ ? 'v1' : $_[1] =~ /3/ ? 'v3' : 'v2c';
-}
+  $self->set($host, $args => [ $oid => OCTET_STRING, $value, ... ], sub {
+    my($self, $err, $res) = @_;
+    # ...
+  });
 
-sub _new_session {
-  my ($self, $args) = @_;
-  my ($session, $error) = Net::SNMP->session(%$args, nonblocking => 1);
+Will call the callback when data is set, instead of emitting the
+L</response> event. C<$args> is optional.
 
-  warn "[SNMP] New session $args->{hostname}: ", ($error || 'OK'), "\n" if DEBUG;
-  $self->emit(error => "$args->{hostname}: $error") if $error;
-  $session;
-}
+=head2 walk
 
-sub _prepare_request {
-  my $self = shift;
-  my $item = shift @{$self->_queue} or return;
-  my ($key, $method, $list, $args, $cb) = @$item;
-  my $session = $self->_pool->{$key};
-  my $success;
+  $self->walk($host, $args, \@oids, sub {
+    my($self, $err, $res) = @_;
+    # ...
+  });
 
-  # dispatch to our mojo based dispatcher
-  $Net::SNMP::DISPATCHER = $self->_dispatcher;
-  warn "[SNMP] >>> $key $method(@$list)\n" if DEBUG;
-
-  Scalar::Util::weaken($self);
-  $method = $SNMP_METHOD{$method} || "$method\_request";
-  $success = $session->$method(
-    $method =~ /bulk/ ? (maxrepetitions => $args->{maxrepetitions} || MAXREPETITIONS) : (),
-    ref $method ? (%$args) : (),
-    varbindlist => $list,
-    callback    => sub {
-      eval {
-        local @$args{qw( method request )} = @$item[1, 2];
-        $self->{_requests}-- if $self->{_requests};
-        if ($_[0]->var_bind_list) {
-          warn "[SNMP] <<< $key $method(@$list)\n" if DEBUG;
-          $cb ? $self->$cb('', $_[0]) : $self->emit(response => $_[0], $args);
-        }
-        else {
-          warn "[SNMP] <<< $key @{[$_[0]->error]}\n" if DEBUG;
-          $cb ? $self->$cb($_[0]->error, undef) : $self->emit(error => $_[0]->error, $_[0], $args);
-        }
-        1;
-      } or do {
-        $self->emit(error => $@);
-      };
-      $self->_prepare_request;
-      $self->_finish unless $self->_dispatcher->connections;
-    },
-  );
-
-  return ++$self->{_requests} if $success;
-  $self->emit(error => $session->error, $session);
-  return $self->{_requests} || '0e0';
-}
-
-sub _finish {
-  warn "[SNMP] Finish\n" if DEBUG;
-  $_[0]->emit('finish');
-  $_[0]->{_setup} = 0;
-}
-
-sub _setup {
-  my $self = shift;
-  my $timeout = $self->master_timeout or return;
-  my $tid;
-
-  warn "[SNMP] Timeout: $timeout\n" if DEBUG;
-  Scalar::Util::weaken($self);
-
-  $tid = $self->ioloop->timer(
-    $timeout => sub {
-      warn "[SNMP] Timeout\n" if DEBUG;
-      $self->ioloop->remove($tid);
-      $self->emit('timeout');
-      $self->{_setup} = 0;
-    }
-  );
-}
+Will call the callback when data is retrieved, instead of emitting the
+L</response> event. C<$args> is optional.
 
 =head2 wait
 
@@ -523,27 +384,196 @@ sub wait {
   $self;
 }
 
-=head2 add_custom_request_method
+for my $method (qw( get get_bulk get_next set walk )) {
+  eval <<"HERE" or die $@;
+    sub $method {
+      my(\$self, \$host) = (shift, shift);
+      my \$args = ref \$_[0] eq 'HASH' ? shift : {};
+      \$self->prepare(\$host, \$args, $method => \@_);
+    }
+    1;
+HERE
+}
 
-  $self->add_custom_request_method(name => sub {
-    my($session, %args) = @_;
-    # do custom stuff..
-  });
+sub _calculate_pool_key {
+  join '|', map { defined $_[1]->{$_} ? $_[1]->{$_} : '' } qw( hostname version community username );
+}
 
-This method can be used to add custom L<Net::SNMP> request methods. See the
-source code for an example on how to do "walk".
+sub _dequeue {
+  my $self = shift;
+  my $key = (keys %{$self->{queue} || {}})[0] or return;
+  return delete $self->{queue}{$key};
+}
 
-NOTE: This method will also replace any method, meaning the code below will
-call the custom callback instead of L<Net::SNMP/get_next_request>.
+sub _finish {
+  warn "[Mojo::SNMP] Finish\n" if DEBUG;
+  $_[0]->emit('finish');
+  $_[0]->{_setup} = 0;
+}
 
-  $self->add_custom_request_method(get_next => $custom_callback);
+sub _new_session {
+  my ($self, $args) = @_;
+  my ($session, $error) = Net::SNMP->new(%$args, nonblocking => 1);
 
-=cut
+  warn "[Mojo::SNMP] New session $args->{hostname}: ", ($error || 'OK'), "\n" if DEBUG;
+  Mojo::IOLoop->next_tick(sub { $self->emit(error => "$args->{hostname}: $error") }) if $error;
+  $session;
+}
 
-sub add_custom_request_method {
-  my ($class, $name, $cb) = @_;
-  $SNMP_METHOD{$name} = $cb;
-  $class;
+sub _normalize_version {
+  $_[1] =~ /1/ ? 'v1' : $_[1] =~ /3/ ? 'v3' : 'v2c';
+}
+
+sub _prepare_request {
+  my ($self, $queue) = @_;
+  my $item = shift @$queue;
+
+  unless ($item) {
+    $queue = $self->_dequeue or return;
+    $item = shift @$queue;
+  }
+
+  my ($key, $method, $list, $args, $cb) = @$item;
+  my $session = $self->{sessions}{$key};
+  my ($error, $success);
+
+  # dispatch to our mojo based dispatcher
+  $Net::SNMP::DISPATCHER = $self->_dispatcher;
+
+  unless ($session->transport) {
+    warn "[Mojo::SNMP] <<< open connection\n" if DEBUG;
+    unless ($session->open) {
+      Mojo::IOLoop->next_tick(
+        sub {
+          return $self->$cb($session->error, undef) if $cb;
+          return $self->emit(error => $session->error, $session, $args);
+        },
+      );
+      return $self->{n_requests} || '0e0';
+    }
+  }
+
+  warn "[Mojo::SNMP] <<< $method $key @$list\n" if DEBUG;
+  Scalar::Util::weaken($self);
+  $method = $SNMP_METHOD{$method} || "$method\_request";
+  $success = $session->$method(
+    $method =~ /bulk/ ? (maxrepetitions => $args->{maxrepetitions} || MAXREPETITIONS) : (),
+    ref $method ? (%$args) : (),
+    varbindlist => $list,
+    callback    => sub {
+      my $session = shift;
+
+      eval {
+        local @$args{qw( method request )} = @$item[1, 2];
+        $self->{n_requests}--;
+        if ($session->var_bind_list) {
+          warn "[Mojo::SNMP] >>> success: $method $key @$list\n" if DEBUG;
+          return $self->$cb('', $session) if $cb;
+          return $self->emit(response => $session, $args);
+        }
+        else {
+          warn "[Mojo::SNMP] >>> error: $method $key @{[$session->error]}\n" if DEBUG;
+          return $self->$cb($session->error, undef) if $cb;
+          return $self->emit(error => $session->error, $session, $args);
+        }
+        1;
+      } or do {
+        $self->emit(error => $@);
+      };
+      warn "[Mojo::SNMP] n_requests: $self->{n_requests}\n" if DEBUG;
+      $self->_prepare_request($queue);
+      warn "[Mojo::SNMP] n_requests: $self->{n_requests}\n" if DEBUG;
+      $self->_finish unless $self->{n_requests};
+    },
+  );
+
+  return ++$self->{n_requests} if $success;
+  $self->emit(error => $session->error, $session);
+  return $self->{n_requests} || '0e0';
+}
+
+sub _setup {
+  my $self = shift;
+  my $timeout = $self->master_timeout or return;
+  my $tid;
+
+  warn "[Mojo::SNMP] Timeout: $timeout\n" if DEBUG;
+  Scalar::Util::weaken($self);
+
+  $tid = $self->ioloop->timer(
+    $timeout => sub {
+      warn "[Mojo::SNMP] Timeout\n" if DEBUG;
+      $self->ioloop->remove($tid);
+      $self->emit('timeout');
+      $self->{_setup} = 0;
+    }
+  );
+}
+
+sub _snmp_method_bulk_walk {
+  my ($session, %args) = @_;
+  my $base_oid       = $args{varbindlist}[0];
+  my $last           = $args{callback};
+  my $maxrepetitions = $args{maxrepetitions} || MAXREPETITIONS;
+  my ($callback, $end, %tree, %types);
+
+  $end = sub {
+    $session->pdu->var_bind_list(\%tree, \%types) if %tree;
+    $session->$last;
+    $end = $callback = undef;
+  };
+
+  $callback = sub {
+    my ($session) = @_;
+    my $res     = $session->var_bind_list    or return $end->();
+    my @sortres = $session->var_bind_names() or return $end->();
+    my $types   = $session->var_bind_types;
+    my $next    = $sortres[-1];
+
+    for my $oid (@sortres) {
+      return $end->() if $types{$oid} or !Net::SNMP::oid_base_match($base_oid, $oid);
+      $types{$oid} = $types->{$oid};
+      $tree{$oid}  = $res->{$oid};
+    }
+
+    return $end->() unless $next;
+    return $session->get_bulk_request(maxrepetitions => $maxrepetitions, varbindlist => [$next], callback => $callback);
+  };
+
+  $session->get_bulk_request(maxrepetitions => $maxrepetitions, varbindlist => [$base_oid], callback => $callback);
+}
+
+sub _snmp_method_walk {
+  my ($session, %args) = @_;
+  my $base_oid = $args{varbindlist}[0];
+  my $last     = $args{callback};
+  my ($callback, $end, %tree, %types);
+
+  $end = sub {
+    $session->pdu->var_bind_list(\%tree, \%types) if %tree;
+    $session->$last;
+    $end = $callback = undef;
+  };
+
+  $callback = sub {
+    my ($session) = @_;
+    my $res = $session->var_bind_list or return $end->();
+    my $types = $session->var_bind_types;
+    my @next;
+
+    for my $oid (keys %$res) {
+      if (!$types{$oid} and Net::SNMP::oid_base_match($base_oid, $oid)) {
+        $types{$oid} = $types->{$oid};
+        $tree{$oid}  = $res->{$oid};
+        push @next, $oid;
+      }
+    }
+
+    return $end->() unless @next;
+    return $session->get_next_request(varbindlist => \@next, callback => $callback);
+  };
+
+  $session->get_next_request(varbindlist => [$base_oid], callback => $callback);
 }
 
 =head1 COPYRIGHT & LICENSE
